import configparser
import json
import os
import socket
import subprocess
from enum import Enum
from itertools import cycle
from pathlib import Path

import requests
import yaml


class RunCode(Enum):
    CONTINUOUS = 0
    ONCE = 1
    UNTIL_SUCCESS = 2
    MONTHLY = 3


class Colors(Enum):
    HEADER = '\033[95m'
    OK_CYAN = '\033[96m'
    OK_GREEN = '\033[92m'
    FAIL = '\033[91m'
    END_C = '\033[0m'


class Probe:

    def __init__(self, detach=True):
        self.script = None
        self.process = None
        if detach:
            self.download()
            self.detach()

    def download(self):
        redirect = requests.get(
            url='%s/download/presto' % os.environ.get('PRELUDE_HQ', 'https://detect.prelude.org'),
            headers=dict(token=os.environ.get('PRELUDE_TOKEN')),
            allow_redirects=False
        )
        if redirect.status_code != 302:
            raise Exception('Failed to download probe: %s' % redirect.status_code)
        self.script = requests.get(redirect.headers['Location']).text

    def detach(self):
        if not self.script:
            raise Exception('No probe available. Please run the "download" function.')
        self.process = subprocess.Popen(['theprelude', '-c', self.script])

    def kill(self):
        if self.process:
            os.kill(self.process.pid, 15)


class Account:

    def __init__(self, profile='default'):
        try:
            parser = configparser.ConfigParser()
            parser.read(os.path.join(Path.home(), '.prelude', 'keychain.ini'))
            profile = next(s for s in parser.sections() if s == profile)
            self.hq = parser.get(profile, 'hq')
            self.secret = parser.get(profile, 'secret')
            self.headers = dict(account=parser.get(profile, 'account'), token=parser.get(profile, 'token'))
        except FileNotFoundError:
            raise Exception('Please create a ~/.prelude/keychain.ini file')
        except KeyError as e:
            raise Exception('Property not found, %s' % e)

    def details(self):
        """ Get Account object """
        res = requests.get('%s/account' % self.hq, headers=self.headers)
        if res.status_code == 200:
            return res.json()
        raise Exception('Failed to get Account (reason:%s)' % res.status_code)

    def update_account(self, key: str, value: str):
        """ Update Account properties """
        properties = dict()
        properties[key] = value
        requests.put('%s/account' % self.hq, headers=self.headers, json=properties)

    def search(self, ttp=None):
        """ Get TTPs matching your query """
        my_outpost = requests.get('%s/account/ttp' % self.hq, headers=self.headers, params=dict(id=ttp))
        if my_outpost.status_code == 200:
            return json.loads(self._xor(my_outpost.text))
        elif my_outpost.status_code == 503:
            raise Exception(my_outpost.text)
        raise Exception('Failed to locate %s. Do you own it?' % ttp)

    def upload(self, path: str):
        """ Save a TTP on your Outpost server """
        with open(path, 'r') as stream:
            procedure = None
            if path.endswith('.yml'):
                procedure = yaml.safe_load(stream)
            procedure = procedure if procedure else json.loads(stream.read())

            encrypted_ttp = self._xor(json.dumps(procedure))
            res = requests.post('%s/account/ttp' % self.hq, headers=self.headers, data=encrypted_ttp)
            if res.status_code != 200:
                raise Exception('Failed to upload: %s' % res.text)

    def activate(self, ttp: str, run_code=RunCode.ONCE):
        """ Activate a TTP so endpoints will start running it """
        res = requests.post(
            url='%s/account/activation/%s' % (self.hq, ttp),
            headers=self.headers,
            json=dict(run_code=run_code.value)
        )
        if res.status_code != 200:
            raise Exception('Failed to activate: %s' % res.text)

    def deactivate(self, ttp: str):
        """ Deactivate a TTP so endpoints will stop running it """
        res = requests.delete('%s/account/activation/%s' % (self.hq, ttp), headers=self.headers)
        if res.status_code != 200:
            raise Exception('Failed to deactivate: %s' % res.text)

    def transfer(self, ttp: str):
        """ Copy Prelude TTPs into Outpost """
        res = requests.get('%s/ttp' % self.hq, headers=self.headers, params=dict(id=ttp))
        if res.status_code == 200:
            encrypted_ttp = self._xor(json.dumps(res.json()))
            res2 = requests.post('%s/account/ttp' % self.hq, headers=self.headers, data=encrypted_ttp)
            if res2.status_code != 200:
                print(f"{Colors.FAIL.value}Failed to transfer: %s, %s{Colors.END_C.value}" % (ttp, res2.text))
            else:
                print(f"{Colors.OK_CYAN.value}Completed transfer of %s{Colors.END_C.value}" % ttp)
        elif res.status_code == 503:
            raise Exception(res.text)
        else:
            raise Exception('Failed to sync procedures: %s' % res.text)

    def register(self, name=socket.gethostname(), tag=None):
        """ Register (or re-register) an endpoint to your account """
        res = requests.post('%s/account/fleet' % self.hq, headers=self.headers, json=dict(id=name, tag=tag))
        if res.status_code == 200:
            return res.text
        raise Exception('Failed to register endpoint: %s' % res.text)

    def activity(self, endpoint_id=None, days=7):
        """ Get a summary of Account activity, or an individual Endpoint """
        params = dict(endpoint_id=endpoint_id, days=days)
        res = requests.get('%s/account/activity' % self.hq, headers=self.headers, params=params)
        if res.status_code == 200:
            return res.json()
        raise Exception('Failed to get activity: %s' % res.text)

    """ PRIVATE """

    def _xor(self, msg):
        return ''.join(chr(ord(c) ^ ord(k)) for c, k in zip(msg, cycle(self.secret)))


class Cli:

    def __init__(self, profile='default'):
        self.account = Account(profile=profile)

    def start(self):
        while True:
            try:
                options = dict(
                    a='Register new device',
                    b='View my fleet',
                    c='Print daily report',
                    d='Activate TTP',
                    e='Deactivate TTP',
                    f='Upload TTP',
                    g='View TTPs',
                    h='Transfer Prelude TTP to my Outpost',
                    i='Update my account'
                )
                selection = input(''.join(['%s) %s\n' % (k, v) for k, v in options.items()]))
                getattr(self, selection)()
            except Exception as ex:
                print(f"{Colors.FAIL.value}%s{Colors.END_C.value}" % ex)

    def a(self):
        name = input(f"{Colors.HEADER.value}Enter a unique name for this device: {Colors.END_C.value}")
        tag = input(f"{Colors.HEADER.value}Enter a tag to categorize: {Colors.END_C.value}")
        token = self.account.register(name=name, tag=tag)
        print(f"{Colors.OK_CYAN.value}Your new token is: %s{Colors.END_C.value}" % token)

    def b(self):
        days = input(f"{Colors.HEADER.value}How many days back do you want to look: {Colors.END_C.value}")
        activity = self.account.activity(days=int(days))
        fleet = []
        for day, results in activity.items():
            for identifier, status in results.items():
                for code, endpoints in status.items():
                    fleet.extend(endpoints)
        probe = input(''.join(['%s) %s\n' % (i, ep) for i, ep in enumerate(set(fleet))]))
        for result in self.account.activity(endpoint_id=fleet[int(probe)], days=int(days)):
            if result['status'] == 0:
                print(f"{Colors.FAIL.value}* %s: %s{Colors.END_C.value}" % (result['created'], result['procedure']))
            elif result['status'] == 1:
                print(f"{Colors.OK_GREEN.value}* %s: %s{Colors.END_C.value}" % (result['created'], result['procedure']))
            else:
                print('* %s: %s' % (result['created'], result['procedure']))

    def c(self):
        days = input(f"{Colors.HEADER.value}How many days back do you want to look: {Colors.END_C.value}")
        expand = input(f"{Colors.HEADER.value}Show results per endpoint? (y/n): {Colors.END_C.value}")
        activity = self.account.activity(days=int(days))
        for day, results in activity.items():
            for identifier, status in results.items():
                success, failure, unknown = [], [], []
                for code, endpoints in status.items():
                    if int(code) == 0:
                        failure = list(set(endpoints))
                    elif int(code) == 1:
                        success = list(set(endpoints))
                    elif int(code) == 2:
                        unknown = list(set(endpoints))
                ttp = self._saturate_ttp(identifier)
                print(f"{Colors.OK_CYAN.value}[%s] %s (%s/%s) {Colors.END_C.value}" % (day, ttp['name'], len(failure), len(success) + len(unknown) + len(failure)))
                if expand == 'y':
                    for endpoint in failure:
                        print(f"{Colors.FAIL.value}* %s{Colors.END_C.value}" % endpoint)
                    for endpoint in success:
                        print(f"{Colors.OK_GREEN.value}* %s{Colors.END_C.value}" % endpoint)
                    for endpoint in unknown:
                        print(f"* N/A: %s" % endpoint)

    def d(self):
        procedures = self.account.search()
        ttp = input(''.join(['%s) %s [%s]\n' % (i, ttp['id'], ttp['name']) for i, ttp in enumerate(procedures)]))
        run_code = input('Select a run code:\n'
                         '0) Continuously\n'
                         '1) Once\n'
                         '2) Until success\n'
                         '3) Monthly\n')
        self.account.activate(procedures[int(ttp)]['id'], run_code=RunCode(int(run_code)))
        print(f"{Colors.OK_GREEN.value}%s has been activated{Colors.END_C.value}" % procedures[int(ttp)]['id'])

    def e(self):
        active = [dict(id=i, code=RunCode(int(rc))) for i, rc in self.account.details().get('active').items()]
        ttp = input(''.join(['%s) %s [%s]\n' % (i, ttp['id'], ttp['code']) for i, ttp in enumerate(active)]))
        self.account.deactivate(active[int(ttp)]['id'])
        print(f"{Colors.OK_GREEN.value}%s has been deactivated{Colors.END_C.value}" % active[int(ttp)]['id'])

    def f(self):
        path = input(f"{Colors.HEADER.value}Enter the absolute path to a TTP file on your computer:: {Colors.END_C.value}")
        self.account.upload(path=path)

    def g(self):
        choose = input('0) Show my Outpost TTPs\n'
                       '1) Show Prelude TTPs\n'
                       '2) I know which identifier I want\n')
        if int(choose) == 0:
            for i in self.account.search():
                print(f"{Colors.OK_CYAN.value}* %s [%s]{Colors.END_C.value}" % (i.get('id'), i.get('name')))
        elif int(choose) == 1:
            res = requests.get('%s/ttp' % self.account.hq, headers=self.account.headers)
            for i in res.json():
                print(f"{Colors.OK_CYAN.value}* %s [%s]{Colors.END_C.value}" % (i.get('id'), i.get('name')))
        elif int(choose) == 2:
            i = input('Enter the identifier: ')
            try:
                ttp = self.account.search(i)
            except Exception:
                ttp = requests.get('%s/ttp' % self.account.hq, headers=self.account.headers, params=dict(id=i)).json()
            if not ttp:
                raise Exception('No TTP by that identifier')
            print(f"{Colors.OK_CYAN.value}%s{Colors.END_C.value}" % json.dumps(ttp, indent=4))

    def h(self):
        procedures = requests.get('%s/ttp' % self.account.hq, headers=self.account.headers).json()
        ttp = input(''.join(['%s) %s (%s)\n' % (i, ttp['id'], ttp['name']) for i, ttp in enumerate(procedures)]))
        self.account.transfer(procedures[int(ttp)]['id'])

    def i(self):
        details = self.account.details()
        print(f"{Colors.HEADER.value}Hit ENTER to keep the existing value{Colors.END_C.value}")
        token = input('Enter a new token (%s): ' % details['token'])
        outpost_address = input('Enter your Outpost FQDN (%s): ' % details['outpost_address'])
        outpost_token = input('Enter your Outpost token (%s): ' % details['outpost_token'])
        if outpost_address:
            self.account.update_account('outpost_address', outpost_address)
        if outpost_token:
            self.account.update_account('outpost_token', outpost_token)
        if token:
            self.account.update_account('token', token)
            print(f"{Colors.FAIL.value}You must update your keychain file and restart!{Colors.END_C.value}")

    def _saturate_ttp(self, i):
        try:
            return self.account.search(ttp=i)
        except Exception:
            return dict(name='Unknown procedure')
